package nachos.threads;

import nachos.machine.*;
import java.util.TreeSet;

/**
 * Uses the hardware timer to provide preemption, and to allow threads to sleep
 * until a certain time.
 */
public class Alarm {
    /**
     * Allocate a new Alarm. Set the machine's timer interrupt handler to this
     * alarm's callback.
     *
     * <p><b>Note</b>: Nachos will not function correctly with more than one
     * alarm.
     */
    public Alarm() {
	Machine.timer().setInterruptHandler(new Runnable() {
		public void run() { timerInterrupt(); }
	    });
    }

    /**
     * The timer interrupt handler. This is called by the machine's timer
     * periodically (approximately every 500 clock ticks). Causes the current
     * thread to yield, forcing a context switch if there is another thread
     * that should be run.
     */
    public void timerInterrupt() {
	long time = Machine.timer().getTime();
	while (!wakeQueue.isEmpty() &&
	       ((WaitingThread) wakeQueue.first()).wakeTime <= time) {
	    WaitingThread w = (WaitingThread) wakeQueue.first();
	    wakeQueue.remove(w);

	    Lib.assert(w.wakeTime <= time);
	    w.thread.ready();
	}
	KThread.currentThread().yield();
    }

    /**
     * Put the current thread to sleep for at least <i>x</i> ticks,
     * waking it up in the timer interrupt handler. The thread must be
     * woken up (placed in the scheduler ready set) during the first timer
     * interrupt where
     *
     * <p><blockquote>
     * (current time) >= (WaitUntil called time)+(x)
     * </blockquote>
     *
     * @param	x	the minimum number of clock ticks to wait.
     *
     * @see	nachos.machine.Timer#getTime()
     */
    public void waitUntil(long x) {
	boolean intStatus = Machine.interrupt().disable();

	wakeQueue.add(new WaitingThread(x));
	KThread.sleep();

	Machine.interrupt().restore(intStatus);
    }
    
    private class WaitingThread implements Comparable {
	WaitingThread(long sleepTime) {
	    thread = KThread.currentThread();
	    wakeTime = Machine.timer().getTime() + sleepTime;
	}

	public int compareTo(Object o) {
	    WaitingThread w = (WaitingThread) o;

	    // can't return 0 for unequal objects, so check all fields
	    if (wakeTime < w.wakeTime)
		return -1;
	    else if (wakeTime > w.wakeTime)
		return 1;
	    else
		return thread.compareTo(w.thread);
	}

	KThread thread;
	long wakeTime;
    }
    
    private TreeSet wakeQueue = new TreeSet();
}
