package nachos.threads;

import nachos.machine.*;

/**
 * A controller for all the elevators in an elevator bank. The controller
 * accesses the elevator bank through an instance of <tt>ElevatorControls</tt>.
 */
public class ElevatorController implements ElevatorControllerInterface {
    /**
     * Allocate a new elevator controller.
     */
    public ElevatorController() {
    }
    
    /**
     * Initialize this elevator controller. The controller will access the
     * elevator bank through <i>controls</i>. This constructor should return
     * immediately after this controller is initialized, but not until the
     * interupt handler is set. The controller will start receiving events
     * after this method returns, but potentially before <tt>run()</tt> is
     * called.
     *
     * @param	controls	the controller's interface to the elevator
     *				bank. The controler must not attempt to access
     *				the elevator bank in <i>any</i> other way.
     */
    public void initialize(ElevatorControls controls) {
        this.controls = controls;

	numFloors = controls.getNumFloors();
	numElevators = controls.getNumElevators();

	lock = new Lock();
	interruptOccurred = new Semaphore(0);

	elevators = new Elevator[numElevators];
	for (int elevator=0; elevator<numElevators; elevator++)
	    elevators[elevator] = new Elevator(elevator);

	upRequests = new boolean[numFloors];
	downRequests = new boolean[numFloors];

	for (int floor=0; floor<numFloors; floor++) {
	    upRequests[floor] = false;
	    downRequests[floor] = false;
	}

	controls.setInterruptHandler(new Runnable() {
	        public void run() { elevatorInterrupt(); }
	    });
    }

    /**
     * Cause the controller to use the provided controls to receive and process
     * requests from riders. This method should not return, but instead should
     * call <tt>controls.finish()</tt> when the controller is finished.
     */
    public void run() {
        for (int elevator=0; elevator<numElevators; elevator++)
	    elevators[elevator].start();

	while (true) {
	    lock.acquire();

	    ElevatorEvent e;
	    int floor, elevator;

	    while ((e = controls.getNextEvent()) != null)
	        handleEvent(e);

	    if (allDone == numElevators)
	        break;

	    lock.release();

	    interruptOccurred.P();
	}

        lock.release();

        for (int elevator=0; elevator<numElevators; elevator++)
	    elevators[elevator].join();

	controls.finish();
    }

    private void elevatorInterrupt() {
      interruptOccurred.V();
    }

    private void handleEvent(ElevatorEvent e) {
        if (e.event == ElevatorEvent.eventRidersDone) {
	    done = true;

	    for (int elevator=0; elevator<numElevators; elevator++)
	        elevators[elevator].wake();
	}
	else if (e.event == ElevatorEvent.eventUpButtonPressed) {
	    upRequests[e.floor] = true;
	    requestChange();
	}
	else if (e.event == ElevatorEvent.eventDownButtonPressed) {
	    downRequests[e.floor] = true;
	    requestChange();
	}
	else if (e.event == ElevatorEvent.eventFloorButtonPressed) {
	    elevators[e.elevator].floorRequest(e.floor);
	}
	else if (e.event == ElevatorEvent.eventElevatorArrived) {
	    elevators[e.elevator].arrived();
	}
	else {
	    Lib.assertNotReached();
	}
    }

    private void requestChange() {
	for (int elevator=0; elevator<numElevators; elevator++)
	    elevators[elevator].adjustDestination();
    }

    private ElevatorControls controls;
    private int numFloors, numElevators;

    private Lock lock;
    private Semaphore interruptOccurred;
  
    private boolean done = false;
    private int allDone = 0;

    private Elevator[] elevators;
    
    private boolean[] upRequests, downRequests;

    private class Elevator implements Runnable {
	Elevator(int elevator) {
	    this.elevator = elevator;

	    floor = floor();
	    destination = floor;
	    
	    thread = new KThread(this);
	    thread.setName("elevator " + elevator);
	    
	    event = new Condition(lock);

	    floorRequests = new boolean[numFloors];
	    
	    for (int floor=0; floor<numFloors; floor++)
		floorRequests[floor] = false;
	}

	void start() {
	    thread.fork();
	}

	public void run() {
	    lock.acquire();

	    while (!done) {
		if (!nextStop()) {
		    sleep();
		    continue;
		}

		if (destination != floor) {
		    int dst = destination;
		    
		    moveTo(dst);
		    
		    while (floor != dst)
			sleep();

		    continue;
		}

		open();
		
		lock.release();
		ThreadedKernel.alarm.waitUntil(timeDoorsOpen);
		lock.acquire();
		
		close();
	    }

	    allDone++;
	    interruptOccurred.V();

	    lock.release();
	}

	void adjustDestination() {
	    nextStop();

	    wake();
	}

	void floorRequest(int floor) {
	    floorRequests[floor] = true;
	    nextStop();

	    wake();
	}

	void arrived() {
	    floor = floor();

	    wake();
	}	    

	private void sleep() {
	    event.sleep();
	}

	void wake() {
	    event.wake();
	}

	void join() {
	    thread.join();
	}

	private boolean nextStop() {
	    // once direction set while the doors are open, no changing allowed
	    if (doorsOpen && dir != dirNeither)
		return false;
		
	    floor = floor();
	    destination = floor;

	    // determine floor increment
	    int incr;
	    if (dir == dirUp)
		incr = 1;
	    else if (dir == dirDown || floor >= numFloors/2)
		incr = -1;
	    else
		incr = 1;
		
	    if (nextStop(incr))
		return true;

	    incr *= -1;

	    if (nextStop(incr))
		return true;

	    incr *= -1;

	    if (nextStop(incr))
		return true;

	    dir = dirNeither;
	    display();
	    return false;
	}
  
	private boolean nextStop(int incr) {
	    // determine which request direction we're searching
	    int requestDir = (incr == 1) ? dirUp : dirDown;
	    boolean[] requests = (incr == 1) ? upRequests : downRequests;

	    while (destination >= 0 && destination < numFloors) {
		if (requests[destination] || floorRequests[destination]) {
		    // set up the direction display
		    if (destination == floor) {
			if (floor == 0 && dir == dirDown ||
			    floor == numFloors-1 && dir == dirUp)
			    dir = dirNeither;
			else
			    dir = requestDir;
		    }
		    else if (destination < floor)
			dir = dirDown;
		    else
			dir = dirUp;

		    display();
		    return true;
		}

		destination += incr;
	    }

	    destination -= incr;
	    return false;
	}

	private int floor() {
	    return controls.getFloor(elevator);
	}

	private boolean moveTo(int destination) {
	    return controls.moveTo(destination, elevator);
	}

	private void clear() {	    
	    floorRequests[floor] = false;

	    if (dir != dirNeither) {
		if (dir == dirUp)
		    upRequests[floor] = false;
		else
		    downRequests[floor] = false;

		requestChange();
	    }
	}

	private void open() {
	    controls.openDoors(elevator);
	    doorsOpen = true;

	    clear();		
	}

	private void close() {
	    doorsOpen = false;
	    controls.closeDoors(elevator);
	}

	private void display() {
	    controls.setDirectionDisplay(elevator, dir);

	    if (doorsOpen)
		clear();
	}
  
	private int elevator, floor, destination;

	private int dir = dirNeither;
	private boolean doorsOpen = false;
  
	private KThread thread;
	private Condition event;
	private boolean[] floorRequests;
    }
}
